﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;

using NUnit.Framework;

using SingularSys.Jep.Functions;
using SingularSys.Jep.Parser;
using SingularSys.Jep.Standard;
using SingularSys.Jep.Types;

namespace SingularSys.Jep.JepTests
{
    [TestFixture]
    public class JepTest
    {
        public const bool PRINT_RESULTS = true;
        /** The parser */
        protected JepInstance jep;

        // @TODO - is this constructor still needed - NUnit is not super-class just annotation
        // public JepTest(String name) : base(name) { }

        /**
         * Sets up the parser.
         */
        [SetUp]
        public void Init()
        {
            // Set up the parser
            jep = new JepInstance();
            jep.ImplicitMul = true;
            //jep.AddStandardFunctions();
            jep.AddStandardConstants();
            //jep.AddComplex();
            //jep.SetTraverse(false);
        }

        /*------------------- utility functions -------------------------------*/

        /**
         * Prints a header with the name of the test as specified in str.
         */
        protected void printTestHeader(String str)
        {
            System.Console.WriteLine("\n\n------------------------------------------------------------------------");
            System.Console.WriteLine("Running \"" + str + "\"\n");
        }

        /**
         * Test result j.evaluate(j.parse(expr))
         * @param expr the expression to parse and evaluate
         * @param expected result expected
         * @throws Exception
         */
        protected void valueTest(String expr, Object expected)
        {
            INode node = jep.Parse(expr);
            Object res = jep.Evaluate(node);

            myAssertEquals(expr, expected, res);
        }

        /**
         * Calculate the value of an expression.
         * @param node
         * @throws Exception
         */
        protected Object calcValue(INode node)
        {
            Object res = jep.Evaluate(node);
            return res;
        }

        /**
         * Calculate the value of an expression.
         * @param expr
         * @throws Exception
         */
        protected Object calcValue(String expr)
        {
            INode node = jep.Parse(expr);
            Object res = calcValue(node);
            return res;
        }

        protected void myAssertEquals(String msg, Object expected, Object act)
        {
            Object actual = null;
            if (act is JepDouble)
                actual = ((JepDouble)act).DoubleValue;
            else
                actual = act;
            if (PRINT_RESULTS && !actual.Equals(expected))
                System.Console.WriteLine("Error: '" + msg + "' is '" + actual + "' should be '" + expected + "'");
            Assert.AreEqual(expected, actual, "<" + msg + ">");
            if (PRINT_RESULTS)
                System.Console.WriteLine("Success: Value of \"" + msg + "\" is " + actual + "");
        }

        public void myAssertNaN(String msg, Object actual)
        {
            if (actual is Double)
            {
                if (Double.IsNaN(((Double)actual)))
                {
                    System.Console.WriteLine("Success: Value of \"" + msg + "\" is " + actual + "");
                }
                else
                {
                    System.Console.WriteLine("Error: \"" + msg + "\" is '" + actual + "' should be NaN");
                    Assert.Fail("<" + msg + "> is " + actual + " should be NaN");
                }
            }
            else
            {
                System.Console.WriteLine("Error: '" + msg + "' is '" + actual + "' should be 'NaN'");
                Assert.Fail("<" + msg + ">");
            }
        }

        /**
         * Test whether an expression evaluates to NaN
         * @param expr
         * @throws Exception
         */
        public void valueTestNaN(String expr)
        {
            Object res = calcValue(expr);
            myAssertNaN(expr, res);
        }

        public void valueTestString(String expr, String expected)
        {
            Object res = calcValue(expr);
            myAssertEquals(expr, expected, res.ToString());
        }
        /** Test parse-evaluate with complex number and given tolerence.
         * 
         * @param expr
         * @param expected
         * @param tol
         * @throws Exception
         */
        public void complexValueTest(String expr, Complex expected, double tol)
        {
            Object res = calcValue(expr);
            if (expected.IsEqual((Complex)res, tol))
                System.Console.WriteLine("Success value of \"" + expr + "\" is " + res);
            else
            {
                System.Console.WriteLine("Error value of \"" + expr + "\" is " + res + " should be " + expected);
                Assert.Fail("<" + expr + "> expected: <" + expected + "> but was <" + res + ">");
            }
        }

        /**
         * Test values to within a given precision
         * 
         * @param expr
         *            expression
         * @param a
         *            the expected value
         * @param tol
         *            tolerance
         * @throws Exception
         */
        public void valueTest(String expr, double a, double tol)
        {
            Object res = calcValue(expr);
            if (res is Double)
            {
                double val = ((Double)res);
                if (Math.Abs(val - a) < tol)
                {
                    System.Console.WriteLine("Success value of \"" + expr + "\" is "
                            + res);
                }
                else
                {
                    System.Console.WriteLine("Error value of \"" + expr + "\" is " + res
                            + " should be " + a);
                    Assert.AreEqual(a, val, tol, expr);
                }
            }
            else
            {
                System.Console.WriteLine("Error value of \"" + expr + "\" is " + res
                        + " should be " + a);
                Assert.Fail("<" + expr + "> expected: <" + a + "> but was <" + res + ">");
            }
        }

        static public void nodeTest(INode n, Operator op)
        {
            Assert.IsTrue(n is ASTOpNode, "Node " + n.ToString() + "should have been an ASTOpNode");
            Assert.AreEqual(op, ((ASTOpNode)n).GetOperator());
        }

        static public void nodeTest(INode n, String name)
        {
            Assert.IsTrue(n is ASTFunNode, "Node " + n.ToString() + "should have been an ASTFunNode");
            Assert.AreEqual(name, ((ASTFunNode)n).GetName());
        }

        static public void nodeTest(INode n, Variable v)
        {
            Assert.IsTrue(n is ASTVarNode, "Node " + n.ToString() + "should have been an ASTVarNode");
            Assert.AreEqual(v, ((ASTVarNode)n).Var);
        }

        static public void nodeTest(INode n, Object v)
        {
            Assert.IsTrue(n is ASTConstant, "Node " + n.ToString() + "should have been an ASTConstant node");
            Assert.AreEqual(v, ((ASTConstant)n).Value);
        }

        public bool compareRecursive(INode node1, INode node2)
        {
            if (node1.JjtGetNumChildren() != node2.JjtGetNumChildren())
                return false;
            if (node1 is ASTConstant)
            {
                if (node2 is ASTConstant)
                {
                    return ((ASTConstant)node1).Value.Equals(((ASTConstant)node2).Value);
                }
                return false;
            }
            if (node1 is ASTVarNode)
            {
                if (node2 is ASTVarNode)
                {
                    return ((ASTVarNode)node1).GetName().Equals(((ASTVarNode)node2).GetName());
                }
                return false;
            }
            if (node1 is ASTOpNode)
            {
                if (node2 is ASTOpNode)
                {
                    if (!((ASTOpNode)node1).GetOperator().Equals(((ASTOpNode)node2).GetOperator()))
                        return false;
                    for (int i = 0; i < node1.JjtGetNumChildren(); ++i)
                    {
                        if (!compareRecursive(node1.JjtGetChild(i), node2.JjtGetChild(i)))
                            return false;
                    }
                    return true;
                }
            }
            if (node1 is ASTFunNode)
            {
                if (node2 is ASTFunNode)
                {
                    if (!((ASTFunNode)node1).GetName().Equals(((ASTFunNode)node2).GetName()))
                        return false;
                    if (((ASTFunNode)node1).GetPFMC() != (((ASTFunNode)node2).GetPFMC()))
                        return false;
                    for (int i = 0; i < node1.JjtGetNumChildren(); ++i)
                    {
                        if (!compareRecursive(node1.JjtGetChild(i), node2.JjtGetChild(i)))
                            return false;
                    }
                    return true;
                }
            }
            return false;
        }

        public void assertEqNodes(String s, INode expected, INode actual)
        {
            bool flag = compareRecursive(expected, actual);
            bool mode = jep.PrintVisitor.GetMode(PrintVisitor.FULL_BRACKET);
            jep.PrintVisitor.SetMode(PrintVisitor.FULL_BRACKET, true);
            if (!flag)
            {
                System.Console.Write("Parse: ");
                System.Console.WriteLine(s.Replace(">\\s+", ">").Replace("\\s+<", "<"));
                System.Console.WriteLine("actual:\t\t" + jep.ToString(actual));
                if (!flag)
                    System.Console.WriteLine("expected:\t" + jep.ToString(expected));
            }
            jep.PrintVisitor.SetMode(PrintVisitor.FULL_BRACKET, mode);
            Assert.IsTrue(flag, "expected: " + jep.ToString(expected) + " actual: " + jep.ToString(actual));
            //		System.Console.WriteLine("\nSuccess");
            //		System.Console.WriteLine(s.replaceAll(">\\s+",">").replaceAll("\\s+<","<"));
            //		jep.println(actual);
        }

        /*------------------------- tests ---------------------------------------*/
        [Test]
        public void testSimpleSum()
        {
            printTestHeader("Testing very simple sums...");

            valueTest("1+2", 3.0);
            valueTest("2*6+3", 15.0);
            valueTest("2*(6+3)", 18.0);
        }

        /**
         * Tests the evaluate() method.
         */
        [Test]
        public void testEvaluate()
        {
            printTestHeader("Testing evaluation results...");
            // test a very basic expression
            Object result = jep.Evaluate(jep.Parse("2.1345"));
            Assert.IsInstanceOfType(typeof(JepDouble), result);
            Assert.AreEqual(2.1345, ((JepDouble)result).DoubleValue);

            // Test Complex numbers
            result = jep.Evaluate(jep.Parse("i"));
            Assert.IsInstanceOfType(typeof(Complex), result);
            Assert.IsTrue(result.Equals(new Complex(0, 1)));

            // Test whether a String is passed through
            result = jep.Evaluate(jep.Parse("\"asdf\""));
            Assert.IsInstanceOfType(typeof(String), result);
            Assert.AreEqual("asdf", result);
        }

        /**
         * Test changing variables value after parsing
         */
        [Test]
        public void testChangeVariable()
        {
            Object result;
            Complex c;
            // add the complex variable x = 0,0
            jep.AddVariable("x", new Complex(0, 0));
            // parse a simple expression with the variable x

            try
            {
                jep.Parse("x");
            }
            catch (ParseException)
            {
                Assert.Fail("Exception occured");
            }
            //		jep.GetVarValue("x");
            try
            {
                result = jep.Evaluate();
            }
            catch (Exception e)
            {
                Assert.Fail("Exception occured" + e.Message);
                return;
            }
            Assert.IsTrue(result is Complex);
            c = (Complex)result;
            Assert.IsTrue(c.Re == 0);
            Assert.IsTrue(c.Im == 0);
            // change the value of x
            jep.AddVariable("x", new Complex(1, 1));
            //		jep.GetVarValue("x");
            try
            {
                result = jep.Evaluate();
            }
            catch (Exception)
            {
                Assert.Fail("Exception occured");
                return;
            }
            Assert.IsTrue(result is Complex);
            c = (Complex)result;
            Assert.IsTrue(c.Re == 1);
            Assert.IsTrue(c.Im == 1);
        }


        /**
         * Tests whether allowUndeclared is working properly. 
         *
         */
        [Test]
        public void testSetAllowUndeclared()
        {
            printTestHeader("Testing AllowedUndeclared options...");

            // test whether setAllowUndeclared(true) works
            jep.VarTab.Clear();				// clear the Variable Table
            jep.AllowUndeclared = true;
            try
            {
                jep.Parse("x");
            }
            catch (ParseException e)
            {
                Assert.Fail("Exception occurred " + e.Message);
            }
            VariableTable st = jep.VarTab;

            // should only contain a single variable x
            Assert.IsTrue(st.Size() == 1);
            Assert.IsTrue(st.GetVariable("x") != null);

            try
            {
                jep.Evaluate();
                Assert.Fail("Exception should have been thrown when trying to evaluate a declared but undefined variable");
            }
            catch (Exception)
            {
                // exception was thrown, so all is well
            }

            jep.SetDefaultValue(0.0);
            jep.Parse("y");
            Object val = jep.Evaluate();
            Assert.AreEqual(0.0, val, "Value of y using default value");

            // test whether setAllowUndeclared(false) works
            jep.VarTab.Clear();
            jep.AddVariable("x", (double)1.0);
            jep.AllowUndeclared = false;
            try
            {
                jep.Parse("p");
                // since p is not declared, an error should occur
                Assert.Fail("A ParseException should have been thrown creating variable 'p'.");
            }
            catch (ParseException)
            {
                // exception was thrown, so all is well
            }
        }

        [Test]
        public void testNumbers()
        {
            printTestHeader("Testing Numbers...");

            valueTest("0", 0.0);
            valueTest("0.", 0.0);
            valueTest(".0", 0.0);
            valueTest("0.0", 0.0);
            valueTest("-0", -0.0);
            valueTest("-0.0", -0.0);
            valueTest("1", 1.0);
            valueTest("1.", 1.0);
            valueTest("1.0", 1.0);
            valueTest("-1", -1.0);
            valueTest("-1.", -1.0);
            valueTest("-1.0", -1.0);
            valueTest("1.5", 1.5);
            valueTest(".5", 0.5);
            valueTest("0.5", 0.5);
            valueTest("-1.5", -1.5);
            valueTest("-.5", -0.5);
            valueTest("-0.5", -0.5);

            valueTest("5e2", 500.0);
            valueTest("5E2", 500.0);
            valueTest("5e-2", 0.05);
            valueTest("5e-2", 0.05);
            valueTest("-5e2", -500.0);
            valueTest("-5E2", -500.0);
            valueTest("-5e-2", -0.05);
            valueTest("-5e-2", -0.05);

            valueTest(".5e3", 500.0);
            valueTest(".5E3", 500.0);
            valueTest(".5e-1", 0.05);
            valueTest(".5e-1", 0.05);
            valueTest("-.5e3", -500.0);
            valueTest("-.5E3", -500.0);
            valueTest("-.5e-1", -0.05);
            valueTest("-.5e-1", -0.05);

            valueTest("0.5e3", 500.0);
            valueTest("0.5E3", 500.0);
            valueTest("0.5e-1", 0.05);
            valueTest(".5e-1", 0.05);
            valueTest("-.5e3", -500.0);
            valueTest("-.5E3", -500.0);
            valueTest("-.5e-1", -0.05);
            valueTest("-.5e-1", -0.05);

            valueTest("1.5e3", 1500.0);
            valueTest("1.5E3", 1500.0);
            valueTest("1.5e-1", 0.15);
            valueTest("1.5e-1", 0.15);
            valueTest("-1.5e3", -1500.0);
            valueTest("-1.5E3", -1500.0);
            valueTest("-1.5e-1", -0.15);
            valueTest("-1.5e-1", -0.15);

        }

        [Test]
        public void testStrings()
        {
            printTestHeader("Testing strings...");

            valueTest("\"\"", "");
            valueTest("\"a\"", "a");
            valueTest("\"abcdefghijklmnopqrstuvwxyz\"", "abcdefghijklmnopqrstuvwxyz");
            valueTest("sum(\"a\",\"b\")", "ab");
            valueTest("\"a\\\"b\"", "a\"b");
            valueTest("\"A\\bB\\fC\\nD\\rE\\tF\"", "A\bB\fC\nD\rE\tF");
            valueTest("A=\"1\"", "1");
            valueTest("B=\"2\"", "2");
            valueTest("(A==\"1\")&&(B==\"2\")", myTrue);
        }

        [Test]
        public void testLogical()
        {
            printTestHeader("Testing logical operations");

            valueTest("T=1", 1.0);
            valueTest("F=0", 0.0);
            valueTest("!T", myFalse);
            valueTest("!F", myTrue);
            valueTest("!5", myFalse);
            valueTest("-0==0", myTrue);
            valueTest("!-5", myFalse);
            //valueTest("-!5==0",myTrue);
            //valueTest("-!0",-1.0);
            valueTest("T&&T", myTrue);
            valueTest("T&&F", myFalse);
            valueTest("F&&T", myFalse);
            valueTest("F&&F", myFalse);
            valueTest("T||T", myTrue);
            valueTest("T||F", myTrue);
            valueTest("F||T", myTrue);
            valueTest("F||F", myFalse);
            calcValue("a=F"); calcValue("b=F"); calcValue("c=F");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=F"); calcValue("b=F"); calcValue("c=T");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=F"); calcValue("b=T"); calcValue("c=F");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=F"); calcValue("b=T"); calcValue("c=T");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);

            calcValue("a=T"); calcValue("b=F"); calcValue("c=F");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=T"); calcValue("b=F"); calcValue("c=T");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=T"); calcValue("b=T"); calcValue("c=F");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);
            calcValue("a=T"); calcValue("b=T"); calcValue("c=T");
            valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))", myTrue);
            valueTest("(a||(b&&c)) == ((a||b)&&(a||c))", myTrue);

            jep.VarTab.Remove("true");
            jep.VarTab.Remove("false");
            jep.AddConstant("true", myTrue);
            jep.AddConstant("false", myFalse);
            valueTest("true", myTrue);
            valueTest("false", myFalse);
            valueTest("!true", myFalse);
            valueTest("!false", myTrue);
            valueTest("true==true", myTrue);
            valueTest("false==false", myTrue);
            valueTest("true==false", myFalse);
            valueTest("true==true&&false==false", myTrue);
            valueTest("if(true==true&&false==false,6,7)", 6.0);
            valueTest("if(false&&true,6,7)", 7.0);
            valueTest("if(true&&false==false,6,7)", 6.0);
            valueTest("if((true&&true)==true,6,7)", 6.0);
            valueTest("if((!false)==true,6,7)", 6.0);
        }

        // @TODO - figure out what to do for Boolean Objects
        Boolean myTrue = true;
        Boolean myFalse = false;
        //protected Object myTrue = Boolean.TRUE; 
        //protected Object myFalse = Boolean.FALSE; 

        [Test]
        public void testNaN()
        {
            printTestHeader("Testing for NaN");
            double x = 1.0, y = 1.0;
            double z = x / y;
            if (z < 2.0)
                System.Console.WriteLine("1/0<2 is true");
            else
                System.Console.WriteLine("1/0<2 is false");

            jep.AddVariable("x", double.NaN);
            System.Console.WriteLine("x=NaN");
            valueTestNaN("ln(x)");
            valueTestNaN("log(x)");
            valueTestNaN("sin(x)");
            valueTestNaN("x+x");
            valueTest("x!=x", myTrue);
            valueTest("x==x", myFalse);

            jep.AddVariable("y", Double.NaN);
            INode n = jep.Parse("x+5");
            System.Console.WriteLine(calcValue(n));
            INode n2 = jep.Parse("y");
            System.Console.WriteLine(calcValue(n2));
            valueTest("x == x+5", myFalse);
            valueTest("x == 0/0", myFalse);
            valueTest("x == x", myFalse);
            valueTest("x == 0 * x", myFalse);
            valueTest("x == 5", myFalse);
            valueTest("x == y", myFalse);
            valueTest("y == y", myFalse);
            System.Console.WriteLine("Set x to Double(5)");
            jep.AddVariable("x", (Double)5);
            valueTest("x == x+5", myFalse);
            valueTest("x == x", myTrue);
        }

        [Test]
        public void testComplex()
        {
            printTestHeader("Testing complex values");
            double tol = 0.00000001;

            complexValueTest("z=complex(3,2)", new Complex(3, 2), tol);
            complexValueTest("z*z-z", new Complex(2, 10), tol);
            complexValueTest("z^3", new Complex(-9, 46), tol);
            complexValueTest("(z*z-z)/z", new Complex(2, 2), tol);
            complexValueTest("w=polar(2,pi/2)", new Complex(0, 2), tol);

            complexValueTest("ln(-1)", new Complex(0, Math.PI), tol);
            complexValueTest("sqrt(-1)", new Complex(0, 1), tol);
            complexValueTest("pow(-1,0.5)", new Complex(0, 1), tol);
            valueTest("arg(w)", Math.PI / 2);
            valueTest("cmod(w)", 2.0);
            valueTest("re(z)", 3.0);
            valueTest("im(z)", 2.0);
            complexValueTest("conj(z)", new Complex(3, -2), tol);
            complexValueTest("exp(pi*i/2)", new Complex(0, 1), tol);
            //complexValueTest("cos(z)",new Complex(3,-2),tol);
        }

        [Test]
        public void testFunction()
        {
            printTestHeader("Testing real functions");
            valueTest("abs(2.5)", 2.5);
            valueTest("abs(-2.5)", 2.5);
            valueTest("acos(1/sqrt(2))", Math.PI / 4, 0.00000001);
            valueTest("cos(pi/3)", 0.5, 0.00000001);
            valueTest("atan2(3,4)", Math.Atan2(3.0, 4.0), 0.00000001);
            valueTest("2^4", 16.0);
            valueTest("2^10", 1024.0);
            valueTest("2^0.5", Math.Sqrt(2), 0.00000001);
            valueTest("2^(-0.5)", 1.0 / Math.Sqrt(2), 0.00000001);
            valueTest("2^(-4)", 1.0 / 16, 0.00000001);
            valueTest("2^(-10)", 1.0 / 1024, 0.00000001);
            calcValue("rand()");
            valueTest("4 atan2(1,1)", Math.PI, 0.00000001);
        }

        [Test]
        public void testIf()
        {
            printTestHeader("Testing if statement");
            valueTest("if(1,2,3)", 2.0);
            valueTest("if(-1,2,3)", 3.0);
            valueTest("if(0,2,3)", 3.0);
            valueTest("if(1,2,3,4)", 2.0);
            valueTest("if(-1,2,3,4)", 3.0);
            valueTest("if(0,2,3,4)", 4.0);
            valueTest("if(0>=0,2,3,4)", 2.0);
            valueTest("x=3", 3.0);
            valueTest("if(x==3,1,-1)", 1.0);
            valueTest("if(x!=3,1,-1)", -1.0);
            valueTest("if(x>=3,1,-1)", 1.0);
            valueTest("if(x>3,1,-1)", -1.0);
            valueTest("if(x<=3,1,-1)", 1.0);
            valueTest("if(x<3,1,-1)", -1.0);
        }

        [Test]
        public void testPlusPlus()
        {
            printTestHeader("++ notation");
            valueTest("3++2", 5.0);
            valueTest("3+-2", 1.0);
            valueTest("3-+2", 1.0);
            valueTest("3--2", 5.0);
            valueTest("3+++2", 5.0);
            valueTest("3++-2", 1.0);
            valueTest("3+-+2", 1.0);
            valueTest("3+--2", 5.0);
            valueTest("3-++2", 1.0);
            valueTest("3-+-2", 5.0);
            valueTest("3--+2", 5.0);
            valueTest("3---2", 1.0);

        }

        [Test]
        public void testImplicitMul()
        {
            printTestHeader("Implicit Multiplication");
            valueTest("x=5", 5.0);
            valueTest("y=7", 7.0);
            valueTest("x2=37", 37.0);
            valueTest("2 x", 10.0);
            valueTest("2x", 10.0);
            valueTest("x 2", 10.0);
            valueTest("x2", 37.0);
            valueTest("-2 x", -10.0);
            valueTest("x y", 35.0);
            valueTest("-x y", -35.0);
            valueTest("x x x x x x x x", Math.Pow(5, 8));
            valueTest("3 (2+4)", 18.0);
            valueTest("3(2+4)", 18.0);
            valueTest("2(x-1)", 8.0);
            valueTest("(x+1)(x-1)", 24.0);
            valueTest("x(x-1)", 20.0);
            valueTest("(x+1)x", 30.0);
            valueTest("1+2 3", 7.0);
            valueTest("2 3+4", 10.0);
            valueTest("1-2 3", -5.0);
            valueTest("2 3 4", 24.0);
            valueTest("-(4-2)", -2.0);
        }

        [Test]
        public void testUminusPower()
        {
            printTestHeader("Testing precedence of unitary minus and power...");
            valueTest("-1^2", -1.0);
            valueTest("2^-3", 0.125);
        }

        [Test]
        public void testNumParam()
        {
            printTestHeader("Number of parameters");
            jep.Parse("if(3,1,2)");
            jep.Parse("if(4,1,2,3)");
            try
            {
                jep.Parse("if(5,1,2,3,4)");
                Assert.Fail("Did not trap illegal number of arguments");
            }
            catch (ParseException) {/* ignore */ }
            jep.Parse("a1=1234");
            jep.Parse("a2=5678");
            jep.Parse("ApportionmentAmt=4321");
            jep.Parse("a4 = 2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + + 1200");
            jep.Parse("a3 = if(a1 > 0 && ApportionmentAmt < 1000, if(a2 < 2000, if(a2 < 1000, 200, 0), if(a1 > 1000, if((2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + 1200 + ApportionmentAmt / 2000 + ApportionmentAmt * ApportionmentAmt + 2000) > 0, 100, 200),200)), if(a1/a2 < 1000, a1/a2, 1, a1 * a2 + a1))");
            try
            {
                jep.Parse("a3 = if(a1 > 0 && ApportionmentAmt < 1000, if(a2 < 2000, if(a2 < 1000, 200, 0), if(a1 > 1000, if((2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + 1200 + ApportionmentAmt / 2000 + ApportionmentAmt * ApportionmentAmt + 2000) > 0, 100, 200)),200), if(a1/a2 < 1000, a1/a2, 1, a1 * a2 + a1))");
                Assert.Fail("Did not trap illegal number of arguments");
            }
            catch (ParseException) { /* ignore */}
        }

        [Test]
        public void testBinom()
        {
            printTestHeader("Testing binomial coeffs");
            valueTest("binom(0,0)", 1);
            valueTest("binom(1,0)", 1);
            valueTest("binom(1,1)", 1);
            valueTest("binom(2,0)", 1);
            valueTest("binom(2,1)", 2);
            valueTest("binom(2,2)", 1);
            valueTest("binom(3,0)", 1);
            valueTest("binom(3,1)", 3);
            valueTest("binom(3,2)", 3);
            valueTest("binom(3,3)", 1);
            valueTest("binom(4,0)", 1);
            valueTest("binom(4,1)", 4);
            valueTest("binom(4,2)", 6);
            valueTest("binom(4,3)", 4);
            valueTest("binom(4,4)", 1);
            valueTest("binom(5,0)", 1);
            valueTest("binom(5,1)", 5);
            valueTest("binom(5,2)", 10);
            valueTest("binom(5,3)", 10);
            valueTest("binom(5,4)", 5);
            valueTest("binom(5,5)", 1);

            valueTest("binom(6,0)", 1);
            valueTest("binom(6,1)", 6);
            valueTest("binom(6,2)", 15);
            valueTest("binom(6,3)", 20);
            valueTest("binom(6,4)", 15);
            valueTest("binom(6,5)", 6);
            valueTest("binom(6,6)", 1);

            valueTest("binom(10,1)", 10);
            valueTest("binom(10,5)", 252);
        }


        //TODO - numberformat stuff needs implementing
        //[Test]
        //public void testFormat() 
        //{
        //    printTestHeader("Format");
        //    NumberFormat format = NumberFormat.getInstance();
        //    jep.PrintVisitor.SetNumberFormat(format);
        //    format.setMaximumFractionDigits(3);
        //    format.setMinimumFractionDigits(0);

        //    String s1 = "[10,0,0.1,0.11,0.111,0.1111]";
        //    String r1 = jep.ToString(jep.Parse(s1));
        //    this.myAssertEquals(s1,"[10,0,0.1,0.11,0.111,0.111]",r1);
        //    String s2 = "[0.9,0.99,0.999,0.9999]";
        //    String r2 = jep.ToString(jep.Parse(s2));
        //    this.myAssertEquals(s2,"[0.9,0.99,0.999,1]",r2);

        //    String s3 = "8 - 7.9"; 
        //    jep.Parse(s3);
        //    double res = jep.EvaluateD();
        //    String s = String.Format("%.3f",res);
        //    this.myAssertEquals(s3,"0.100",s);

        //    String s4 = "round(8-7.9,3)";
        //    jep.Parse(s4);
        //    res = jep.EvaluateD();
        //    this.myAssertEquals(s4,"0.1",String.valueOf(res));
        //}

        [Test]
        public void testAssign()
        {
            printTestHeader("Assignment of variables");
            valueTest("x=3", 3.0);
            valueTest("y=3+4", 7.0);
            valueTest("z=x+y", 10.0);
            valueTest("a=b=c=z", 10.0);
            valueTest("b", 10.0);
            valueTest("d=f=a-b", 0.0);
            valueTest("x=2", 2.0);
            valueTest("(x*x)*x*(x*x)", 32.0); // Works fine with Multiply
            //		new org.lsmp.djep.vectorJep.VectorJep();
            valueTest("(x*x)*x*(x*x)", 32.0);
            // this created an error in 2.3.0b
            // as creating a VectorJep changed the operator set
            // and hence the broken MMultiply was used.								
        }

        [Test]
        public void testNoAssign()
        {
            printTestHeader("Assignment of variables");
            jep.AllowAssignment = false;
            try
            {
                jep.Parse("x=3");
                Assert.Fail("Assignment should not have been possible");
            }
            catch (ParseException)
            {
                System.Console.WriteLine("Attept at assignment sucessfully caught");
            }
            jep.AllowAssignment = true;
            jep.Parse("x=3");
        }

        [Test]
        public void testListAccess()
        {
            printTestHeader("List operations");
            valueTestString("x=[4,3,2,1]", "[4.0, 3.0, 2.0, 1.0]");
            valueTest("x[2]", 3.0);
            valueTest("x[4]=5", 5.0);
            valueTestString("x", "[4.0, 3.0, 2.0, 5.0]");
            valueTestString("[]", "[]");
            valueTestString("x+x", "[8.0, 6.0, 4.0, 10.0]");
            valueTestString("x-x", "[0.0, 0.0, 0.0, 0.0]");
            valueTestString("x*2", "[8.0, 6.0, 4.0, 10.0]");
            valueTestString("2*x", "[8.0, 6.0, 4.0, 10.0]");
            valueTestString("x/2", "[2.0, 1.5, 1.0, 2.5]");
            valueTestString("w=[1+i,1-i]", "[(1.0, 1.0), (1.0, -1.0)]");
            valueTestString("2*w", "[(2.0, 2.0), (2.0, -2.0)]");
        }

        [Test]
        public void testMultiDimArray()
        {
            printTestHeader("List operations");
            valueTestString("x=[[1,2],[3,4]]", "[[1.0, 2.0], [3.0, 4.0]]");
            valueTestString("x[2]", "[3.0, 4.0]");
            valueTestString("x+x", "[[2.0, 4.0], [6.0, 8.0]]");
            valueTestString("x-x", "[[0.0, 0.0], [0.0, 0.0]]");
            valueTestString("x*2", "[[2.0, 4.0], [6.0, 8.0]]");
            valueTestString("2*x", "[[2.0, 4.0], [6.0, 8.0]]");
            valueTestString("x/2", "[[0.5, 1.0], [1.5, 2.0]]");
        }


        class SideEffect : PostfixMathCommand
        {
            private static readonly long serialVersionUID = 1L;
            public SideEffect() { this.SetCurNumberOfParameters(1); }
            public bool called = false;
            public void run(Stack<Object> aStack)
            {
                called = true;
            }
        }

        [Test]
        public void testLazyLogical()
        {
            printTestHeader("Lazy Logical");
            // TODO - add LazyLogical to /Functions
            jep.OpTab.GetAnd().SetPFMC(new LazyLogical(LazyLogical.AND));
            jep.OpTab.GetOr().SetPFMC(new LazyLogical(LazyLogical.OR));

            SideEffect se = new SideEffect();
            jep.AddFunction("sideEffect", se);
            INode n1 = jep.Parse("0 || sideEffect(1==2)");
            Object v1 = jep.Evaluate(n1);
            Assert.IsTrue(se.called);
            Assert.AreEqual(myFalse, v1);
            se.called = false;

            INode n2 = jep.Parse("1 || sideEffect(1==2)");
            Object v2 = jep.Evaluate(n2);
            Assert.IsFalse(se.called);
            Assert.AreEqual(myTrue, v2);
            se.called = false;

            INode n3 = jep.Parse("0 && sideEffect(1==2)");
            Object v3 = jep.Evaluate(n3);
            Assert.IsFalse(se.called);
            Assert.AreEqual(myFalse, v3);
            se.called = false;

            INode n4 = jep.Parse("1 && sideEffect(1==2)");
            Object v4 = jep.Evaluate(n4);
            Assert.IsTrue(se.called);
            Assert.AreEqual(myFalse, v4);
            se.called = false;

            INode n5 = jep.Parse("0 || sideEffect(2==2)");
            Object v5 = jep.Evaluate(n5);
            Assert.IsTrue(se.called);
            Assert.AreEqual(myTrue, v5);
            se.called = false;

            INode n6 = jep.Parse("1 || sideEffect(2==2)");
            Object v6 = jep.Evaluate(n6);
            Assert.IsFalse(se.called);
            Assert.AreEqual(myTrue, v6);
            se.called = false;

            INode n7 = jep.Parse("0 && sideEffect(2==2)");
            Object v7 = jep.Evaluate(n7);
            Assert.IsFalse(se.called);
            Assert.AreEqual(myFalse, v7);
            se.called = false;

            INode n8 = jep.Parse("1 && sideEffect(2==2)");
            Object v8 = jep.Evaluate(n8);
            Assert.IsTrue(se.called);
            Assert.AreEqual(myTrue, v8);
            se.called = false;

            INode n9 = jep.Parse("if(1,sideEffect(2),3)");
            Object v9 = jep.Evaluate(n9);
            Assert.IsTrue(se.called);
            Assert.AreEqual((Double)2, v9);
            se.called = false;

            INode n10 = jep.Parse("if(1,2,sideEffect(3))");
            Object v10 = jep.Evaluate(n10);
            Assert.IsFalse(se.called);
            Assert.AreEqual((Double)2, v10);
            se.called = false;

            INode n11 = jep.Parse("if(0,sideEffect(2),3)");
            Object v11 = jep.Evaluate(n11);
            Assert.IsFalse(se.called);
            Assert.AreEqual((Double)3, v11);
            se.called = false;

            INode n12 = jep.Parse("if(0,2,sideEffect(3))");
            Object v12 = jep.Evaluate(n12);
            Assert.IsTrue(se.called);
            Assert.AreEqual((Double)3, v12);
            se.called = false;
        }

        // TODO - add StrictNaturalLogarithm to /Functions
        //[Test]
        //public void testLogarithm() 
        //{
        //    printTestHeader("Logarithm");
        //    FunctionTable ft = jep.FunTab;
        //    ft.AddFunction("slog",new StrictNaturalLogarithm());
        //    ft.AddFunction("slog10",new StrictNaturalLogarithm(10));
        //    ft.AddFunction("slog2",new StrictNaturalLogarithm(2));
        //    jep.AddVariable("NaN",Double.NaN);
        //    valueTest("ln(1)",0.0,1E-7);
        //    valueTest("slog(1)",0.0,1E-7);
        //    valueTest("slog10(1)",0.0,1E-7);
        //    valueTest("ln(e)",1.0,1E-7);
        //    valueTest("slog(e)",1.0,1E-7);
        //    valueTest("slog10(10)",1.0,1E-7);
        //    valueTestNaN("slog(-1)");
        //    valueTestNaN("slog10(-1)");
        //    valueTestNaN("slog2(-1)");
        //}

        /**
         * Test creating a Jep instance with no functions and no variables.
         * @throws Exception
         */
        [Test]
        public void testBlankParser()
        {
            printTestHeader("Testing blank parser");
            JepInstance j = new JepInstance();
            // override standard variable table with blank one
            j.VarTab = new VariableTable();
            // override standard function table with blank one
            j.FunTab = new FunctionTable();

            // ensure function and variable tables are blank
            Assert.IsTrue(j.FunTab.IsEmpty(), "Function table not empty");
            Assert.IsTrue(j.VarTab.IsEmpty(), "Variable table not empty");
            // ensure operators still work
            j.Parse("1+1");
            Object res = j.Evaluate();
            Assert.IsInstanceOfType(typeof(JepDouble), res);
            Assert.AreEqual(2.0, ((JepDouble)res).DoubleValue);
        }

        // TODO - add MacroFunction to /Misc
        //[Test]
        //public void testMacroFunction() 
        //{
        //    printTestHeader("Testing MacroFunctions...");
        //    MacroFunction fact = new MacroFunction(
        //        "fact",new String[]{"x"}, 
        //        "if(x>1,x*fact(x-1),1)");
        //    jep.AddFunction("fact",fact);
        //    fact.init(jep);
        //    valueTest("fact(5)",120.0);
        //    jep.AddVariable("x",4);
        //    valueTest("fact(5)",120.0);
        //    Assert.AreEqual(4.0,jep.GetVariableValue("x"));
        //    valueTest("x+fact(x+1)",124.0);

        //}

        /**
         * Tests whether a parse exceptions are thrown when they should be. 
         * See ticket #50
         * Tests
         * on both the standard parser and the configurable parser, is carried out
         * by CPTest which extends this.
         */
        [Test]
        public void testParseException()
        {
            printTestHeader("Testing ParseException...");

            // all these expressions should cause ParseException
            String[] expr = { "1+", "(", ")", "sin(", "(1+)", "*1", "1*", "-", "avg()", "min()", "max()" };

            // for each expression
            for (int i = 0; i < expr.Length; i++)
            {
                try
                {
                    System.Console.WriteLine("Attempting parse of \"" + expr[i] + "\"");
                    jep.Parse(expr[i]);
                    // expression should have failed to parse
                    Assert.Fail("Attempting parse of \"" + expr[i] + "\"");
                }
                catch (ParseException e)
                {
                    // expected a parse exception has been thrown
                    System.Console.WriteLine("Caught expected exception \"" + e.Message + "\"");
                }
                catch (Exception e)
                {
                    // some other exception was thrown
                    Assert.Fail("Attempting parse of \"" + expr[i] + "\"" + e.Message);
                }
            }
        }

        /**
         * Test whether evaluation exceptions are thrown when they should be.
         */
        [Test]
        public void testEvalExceptions()
        {
            printTestHeader("Ensuring evaluation exception are thrown when they should be...");
            ArrayList exprs = new ArrayList();

            // random expressions that will fail during evaluation
            //exprs.Add("avg()");
            //exprs.Add("min()");
            //exprs.Add("max()");
            exprs.Add("avg([])");
            exprs.Add("min([])");
            exprs.Add("max([])");

            // add expressions like sin([1, 1]) that can not handle arrays
            String[] noArrayFuncs = { // list of functions that can not take arrays as arguments
				    "sin", "cos", "tan", "asin", "acos", "atan", 
				    "sinh", "cosh", "tanh", "asinh", "acosh", "atanh",
				    "log", "ln", "exp"
		    };
            foreach (String funcname in noArrayFuncs)
            {
                exprs.Add(funcname + "([])");
                exprs.Add(funcname + "([1,1])");
            }

            // loop through each exception
            foreach (String str in exprs)
            {
                // parsing should work fine
                try
                {
                    jep.Parse(str);
                }
                catch (ParseException e)
                {
                    Assert.Fail("Could not parse expression: \"" + str + "\", " + e.Message);
                }
                // but evaluation should fail
                try
                {
                    jep.Evaluate();
                    // we shouldn't get here because these expressions should
                    // end up throwing eval expections
                    Assert.Fail("No evaluation exception was thrown as expected.");
                }
                catch (Exception)
                {
                    // we should end up here
                }
            }
        }

        [Test]
        public void testSemiColon()
        {
            String s = "x=5; y=6; x+y";
            printTestHeader("Testing reading multiple equations: " + s);
            jep.InitMultiParse(s);
            INode n1 = jep.ContinueParsing();
            INode n2 = jep.ContinueParsing();
            INode n3 = jep.ContinueParsing();
            Object o1 = jep.Evaluate(n1);
            this.myAssertEquals("x=5", 5.0, o1);
            Object o2 = jep.Evaluate(n2);
            this.myAssertEquals("y=6", 6.0, o2);
            Object o3 = jep.Evaluate(n3);
            this.myAssertEquals("x+y", 11.0, o3);
            INode n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");
        }

        [Test]
        public void testMultiLine()
        {
            String s = "\nx=\n5;y=6;\nx+y";
            printTestHeader("Testing reading multiple equations: " + s);
            jep.InitMultiParse(s);

            INode n1 = jep.ContinueParsing();
            Object o1 = jep.Evaluate(n1);
            this.myAssertEquals("x=5", 5.0, o1);

            INode n2 = jep.ContinueParsing();
            Object o2 = jep.Evaluate(n2);
            this.myAssertEquals("y=6", 6.0, o2);

            INode n3 = jep.ContinueParsing();
            Object o3 = jep.Evaluate(n3);
            this.myAssertEquals("x+y", 11.0, o3);

            INode n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

            /******/

            s = "x*y";
            jep.InitMultiParse(s);

            n1 = jep.ContinueParsing();
            o1 = jep.Evaluate(n1);
            myAssertEquals("x*y", 30.0, o1);

            n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

            /******/

            s = "\n";
            jep.InitMultiParse(s);
            n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

            /******/

            s = "";
            jep.InitMultiParse(s);
            n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

            /******/

            s = "\n\n\n";
            jep.InitMultiParse(s);
            n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

            /******/

            s = "\n\n\n1\n\n\n";
            jep.InitMultiParse(s);
            n1 = jep.ContinueParsing();
            o1 = jep.Evaluate(n1);
            myAssertEquals("1", 1.0, o1);
            n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");

        }

        [Test]
        public void testEmptyEqn()
        {
            String s = "x=5;;y=6;x+y";
            printTestHeader("Testing reading multiple equations: " + s);
            jep.InitMultiParse(s);

            INode n1 = jep.ContinueParsing();
            Object o1 = jep.Evaluate(n1);
            this.myAssertEquals("x=5", 5.0, o1);

            INode n2 = jep.ContinueParsing();
            Object o2 = jep.Evaluate(n2);
            this.myAssertEquals("y=6", 6.0, o2);

            INode n3 = jep.ContinueParsing();
            Object o3 = jep.Evaluate(n3);
            this.myAssertEquals("x+y", 11.0, o3);

            INode n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");
        }

        [Test]
        public void testComments()
        {
            printTestHeader("Testing comments: ");
            NodeFactory nf = jep.NodeFac;
            OperatorTable ot = jep.OpTab;

            INode expected =
            nf.BuildOperatorNode(
                ot.GetDivide(),
                nf.BuildOperatorNode(
                    ot.GetMultiply(),
                    nf.BuildVariableNode("x"),
                    nf.BuildVariableNode("y")),
                nf.BuildVariableNode("z"));

            String s = "x*y/z //foo\n";
            INode node = jep.Parse(s);
            System.Console.WriteLine(node.ToString());
            jep.PrintLine();
            this.assertEqNodes(s, expected, node);

            s = "//foo\nx*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x//foo\n*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* foo */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* foo */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/ /* foo */z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/* foo *//z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*/* foo */y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x/* foo */*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "/* foo */x*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);



            s = "x*y/z /* foo/bar */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z/* foo*bar */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);


            s = "x*y/z/* foo\n */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/ /* foo\n */z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/* foo\n *//z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*/* foo\n */y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x/* foo\n */*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "/* foo\n */x*y/z";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* foo/bar\n */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* foo*bar\n */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);


            s = "x*y/z /* \nfoo/bar */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* \nfoo*bar */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* \nfoo/bar\n */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

            s = "x*y/z /* \nfoo*bar\n */";
            node = jep.Parse(s);
            this.assertEqNodes(s, expected, node);

        }

        [Test]
        public void testEmptyEqn2()
        {
            String s = "x=5;/*foo*/;y=6;x+y";
            printTestHeader("Testing reading multiple equations: " + s);
            jep.InitMultiParse(s);

            INode n1 = jep.ContinueParsing();
            Object o1 = jep.Evaluate(n1);
            this.myAssertEquals("x=5", 5.0, o1);

            INode n2 = jep.ContinueParsing();
            Object o2 = jep.Evaluate(n2);
            this.myAssertEquals("y=6", 6.0, o2);

            INode n3 = jep.ContinueParsing();
            Object o3 = jep.Evaluate(n3);
            this.myAssertEquals("x+y", 11.0, o3);

            INode n4 = jep.ContinueParsing();
            Assert.IsNull(n4, "Next parse should give a null result");
        }

        [Test]
        public void testX2Y()
        {
            String s = "x^2y";
            INode node = jep.Parse(s);
            NodeFactory nf = jep.NodeFac;
            OperatorTable ot = jep.OpTab;
            INode expected =
            nf.BuildOperatorNode(
                ot.GetMultiply(),
                nf.BuildOperatorNode(
                    ot.GetPower(),
                    nf.BuildVariableNode("x"),
                    nf.BuildConstantNode(2.0)),
                nf.BuildVariableNode("y"));
            this.assertEqNodes(s, expected, node);

            s = "!x y";
            node = jep.Parse(s);
            expected =
            nf.BuildOperatorNode(
                ot.GetMultiply(),
                nf.BuildOperatorNode(
                    ot.GetNot(),
                    nf.BuildVariableNode("x")),
                nf.BuildVariableNode("y"));
            this.assertEqNodes(s, expected, node);

            s = "x*2y";
            node = jep.Parse(s);
            expected =
            nf.BuildOperatorNode(
                ot.GetMultiply(),
                nf.BuildOperatorNode(
                    ot.GetMultiply(),
                    nf.BuildVariableNode("x"),
                    nf.BuildConstantNode(2.0)),
                nf.BuildVariableNode("y"));
            this.assertEqNodes(s, expected, node);

            s = "x/2y";
            node = jep.Parse(s);
            expected =
            nf.BuildOperatorNode(
                ot.GetMultiply(),
                nf.BuildOperatorNode(
                    ot.GetDivide(),
                    nf.BuildVariableNode("x"),
                    nf.BuildConstantNode(2.0)),
                nf.BuildVariableNode("y"));
            this.assertEqNodes(s, expected, node);

            s = "x+2y";
            node = jep.Parse(s);
            expected =
            nf.BuildOperatorNode(
                ot.GetAdd(),
                nf.BuildVariableNode("x"),
                nf.BuildOperatorNode(
                    ot.GetMultiply(),
                    nf.BuildConstantNode(2.0),
                    nf.BuildVariableNode("y")));
            this.assertEqNodes(s, expected, node);

        }

        /**
         * Tests for bug #52
         */
        [Test]
        public void testExpectedTokenSequence()
        {
            printTestHeader("Checking whether expected token sequence is set");
            JepInstance j = new JepInstance();
            try
            {
                j.Parse("1+");
            }
            catch (SingularSys.Jep.ParseException e)
            {
                // TODO verify this implementation
                //Throwable t = e.getCause();
                Exception t = e.GetBaseException();

                // ensure the cause is set
                Assert.IsNotNull(t);
                if (t is SingularSys.Jep.ParseException)
                {
                    SingularSys.Jep.Parser.ParseException specialE =
                        (SingularSys.Jep.Parser.ParseException)t;
                    System.Console.WriteLine("Contains parser.ParseException :)");
                    int[][] expectedTokens = specialE.expectedTokenSequences;
                    Assert.IsTrue(expectedTokens.Length > 0);
                }
                else
                {
                    Assert.Fail();
                }
            }
        }

        // Test for bug with 1&&1&&1
        [Test]
        public void testLazyLogicalBug()
        {
            valueTest("true&&1", myTrue);
        }
    }
}


#region ORIGINAL JAVA CODE
//package com.singularsys.jeptests;

//import static org.junit.Assert.assertTrue;

//import java.text.NumberFormat;
//import java.util.ArrayList;
//import java.util.Stack;

//import com.singularsys.jep.*;
//import com.singularsys.jep.configurableparser.ConfigurableParser;
//import com.singularsys.jep.configurableparser.StandardConfigurableParser;
//import com.singularsys.jep.functions.LazyLogical;
//import com.singularsys.jep.functions.PostfixMathCommand;
//import com.singularsys.jep.functions.StrictNaturalLogarithm;
//import com.singularsys.jep.misc.MacroFunction;
//import com.singularsys.jep.parser.ASTConstant;
//import com.singularsys.jep.parser.ASTFunNode;
//import com.singularsys.jep.parser.ASTOpNode;
//import com.singularsys.jep.parser.ASTVarNode;
//import com.singularsys.jep.parser.Node;
//import com.singularsys.jep.standard.Complex;

//import junit.framework.Assert;
//import junit.framework.TestCase;

//public class JepTest extends TestCase {
//    public static final boolean PRINT_RESULTS = true;
//    /** The parser */
//    protected Jep jep;
		
//    public JepTest(String name) {
//        super(name);
//    }
	
//    /**
//     * Sets up the parser.
//     */
//    @Override
//    public void setUp() {
//        // Set up the parser
//        jep = new Jep();
//        jep.setImplicitMul(true);
//        //jep.addStandardFunctions();
//        jep.addStandardConstants();
//        //jep.addComplex();
//        //jep.setTraverse(false);
//    }
	
//    /*------------------- utility functions -------------------------------*/
	
//    /**
//     * Prints a header with the name of the test as specified in str.
//     */
//    protected void printTestHeader(String str) {
//        System.out.println("\n\n------------------------------------------------------------------------");
//        System.out.println("Running \""+str+"\"\n");
//    }
	
//    /**
//     * Test result j.evaluate(j.parse(expr))
//     * @param expr the expression to parse and evaluate
//     * @param expected result expected
//     * @throws Exception
//     */
//    protected void valueTest(String expr,Object expected) throws Exception
//    {
//        Node node = jep.parse(expr);
//        Object res = jep.evaluate(node);
		
//        myAssertEquals(expr,expected,res);
//    }

//    /**
//     * Calculate the value of an expression.
//     * @param node
//     * @throws Exception
//     */
//    protected Object calcValue(Node node) throws Exception {
//        Object res = jep.evaluate(node);
//        return res;
//    }

//    /**
//     * Calculate the value of an expression.
//     * @param expr
//     * @throws Exception
//     */
//    protected Object calcValue(String expr) throws Exception {
//        Node node = jep.parse(expr);
//        Object res = calcValue(node);
//        return res;
//    }

//    protected void myAssertEquals(String msg, Object expected, Object actual) {
//        if(PRINT_RESULTS && !actual.equals(expected))
//            System.out.println("Error: '"+msg+"' is '"+actual+"' should be '"+expected+"'");
//        assertEquals("<"+msg+">",expected,actual);
//        if(PRINT_RESULTS)
//            System.out.println("Success: Value of \""+msg+"\" is "+actual+"");
//    }

//    public void myAssertNaN(String msg,Object actual)
//    {
//        if(actual instanceof Double) {
//            if(Double.isNaN( ((Double) actual).doubleValue()) ) {
//                System.out.println("Success: Value of \""+msg+"\" is "+actual+"");
//            }
//            else {
//                System.out.println("Error: \""+msg+"\" is '"+actual+"' should be NaN");
//                fail("<"+msg+"> is "+actual+" should be NaN");
//            }
//        }
//        else {
//            System.out.println("Error: '"+msg+"' is '"+actual+"' should be 'NaN'");
//            fail("<"+msg+">");
//        }
//    }

//    /**
//     * Test whether an expression evaluates to NaN
//     * @param expr
//     * @throws Exception
//     */
//    public void valueTestNaN(String expr) throws Exception
//    {
//        Object res = calcValue(expr);
//        myAssertNaN(expr,res);
//    }

//    public void valueTestString(String expr,String expected) throws Exception
//    {
//        Object res = calcValue(expr);
//        myAssertEquals(expr,expected,res.toString());
//    }
//    /** Test parse-evaluate with complex number and given tolerence.
//     * 
//     * @param expr
//     * @param expected
//     * @param tol
//     * @throws Exception
//     */
//    public void complexValueTest(String expr,Complex expected,double tol) throws Exception
//    {
//        Object res = calcValue(expr);
//        if(expected.equals((Complex) res,tol))
//            System.out.println("Success value of \""+expr+"\" is "+res);
//        else {
//            System.out.println("Error value of \""+expr+"\" is "+res+" should be "+expected);
//            fail("<"+expr+"> expected: <"+expected+"> but was <"+res+">");
//        }
//    }

//    /**
//     * Test values to within a given precision
//     * 
//     * @param expr
//     *            expression
//     * @param a
//     *            the expected value
//     * @param tol
//     *            tolerance
//     * @throws Exception
//     */
//    public void valueTest(String expr, double a, double tol) throws Exception {
//        Object res = calcValue(expr);
//        if (res instanceof Double) {
//            double val = ((Double) res).doubleValue();
//            if (Math.abs(val - a) < tol) {
//                System.out.println("Success value of \"" + expr + "\" is "
//                        + res);
//            } else {
//                System.out.println("Error value of \"" + expr + "\" is " + res
//                        + " should be " + a);
//                assertEquals(expr, a, val, tol);
//            }
//        } else {
//            System.out.println("Error value of \"" + expr + "\" is " + res
//                    + " should be " + a);
//            fail("<" + expr + "> expected: <" + a + "> but was <" + res + ">");
//        }
//    }

//    static public void nodeTest(Node n, Operator op) {
//        assertTrue("Node " + n.toString() + "should have been an ASTOpNode",
//                n instanceof ASTOpNode);
//        assertEquals(op, ((ASTOpNode) n).getOperator());
//    }

//    static public void nodeTest(Node n, String name) {
//        assertTrue("Node " + n.toString() + "should have been an ASTFunNode",
//                n instanceof ASTFunNode);
//        assertEquals(name, ((ASTFunNode) n).getName());
//    }

//    static public void nodeTest(Node n, Variable v) {
//        assertTrue("Node " + n.toString() + "should have been an ASTVarNode",
//                n instanceof ASTVarNode);
//        assertEquals(v, ((ASTVarNode) n).getVar());
//    }

//    static public void nodeTest(Node n, Object v) {
//        assertTrue("Node " + n.toString() + "should have been an ASTConstant node",
//                n instanceof ASTConstant);
//        assertEquals(v, ((ASTConstant) n).getValue());
//    }

//        public boolean compareRecursive(Node node1,Node node2) {
//        if(node1.jjtGetNumChildren()!=node2.jjtGetNumChildren())
//            return false;
//        if(node1 instanceof ASTConstant) {
//            if(node2 instanceof ASTConstant) {
//            return ((ASTConstant)node1).getValue().equals(((ASTConstant)node2).getValue());
//            }
//            return false;
//        }
//        if(node1 instanceof ASTVarNode) {
//            if(node2 instanceof ASTVarNode) {
//            return ((ASTVarNode) node1).getName().equals(((ASTVarNode) node2).getName());
//            }
//            return false;
//        }
//        if(node1 instanceof ASTOpNode) {
//            if(node2 instanceof ASTOpNode) {
//            if(!((ASTOpNode) node1).getOperator().equals(((ASTOpNode) node2).getOperator()))
//                return false;
//            for(int i=0;i<node1.jjtGetNumChildren();++i) {
//                if(!compareRecursive(node1.jjtGetChild(i),node2.jjtGetChild(i)))
//                    return false;
//            }
//            return true;
//            }
//        }
//        if(node1 instanceof ASTFunNode) {
//            if(node2 instanceof ASTFunNode) {
//            if(!((ASTFunNode) node1).getName().equals(((ASTFunNode) node2).getName()))
//                return false;
//            if(((ASTFunNode) node1).getPFMC() != (((ASTFunNode) node2).getPFMC()))
//                return false;
//            for(int i=0;i<node1.jjtGetNumChildren();++i) {
//                if(!compareRecursive(node1.jjtGetChild(i),node2.jjtGetChild(i)))
//                    return false;
//            }
//            return true;
//            }
//        }
//        return false;
//        }

//        public void assertEqNodes(String s,Node expected,Node actual) {
//        boolean flag = compareRecursive(expected,actual);
//        boolean mode = jep.getPrintVisitor().getMode(PrintVisitor.FULL_BRACKET);
//        jep.getPrintVisitor().setMode(PrintVisitor.FULL_BRACKET,true);
//        if(!flag ) {
//            System.out.print("Parse: ");
//            System.out.println(s.replaceAll(">\\s+",">").replaceAll("\\s+<","<"));
//            System.out.println("actual:\t\t"+jep.toString(actual));
//            if(!flag)
//            System.out.println("expected:\t"+jep.toString(expected));
//        }
//        jep.getPrintVisitor().setMode(PrintVisitor.FULL_BRACKET,mode);
//        assertTrue("expected: "+jep.toString(expected)+" actual: "+jep.toString(actual),
//            flag);
////		System.out.println("\nSuccess");
////		System.out.println(s.replaceAll(">\\s+",">").replaceAll("\\s+<","<"));
////		jep.println(actual);
//        }

//    /*------------------------- tests ---------------------------------------*/

//    public void testSimpleSum() throws Exception
//    {
//        printTestHeader("Testing very simple sums...");

//        valueTest("1+2",3.0);		
//        valueTest("2*6+3",15.0);		
//        valueTest("2*(6+3)",18.0);
//    }

//    /**
//     * Tests the evaluate() method.
//     */
//    public void testEvaluate() throws Exception {
//        printTestHeader("Testing evaluation results...");
//            // test a very basic expression
//            Object result = jep.evaluate(jep.parse("2.1345"));
//            Assert.assertTrue(result instanceof Double);
//            Assert.assertTrue(result.equals(2.1345));

//            // Test Complex numbers
//            result = jep.evaluate(jep.parse("i"));
//            Assert.assertTrue(result instanceof Complex);
//            Assert.assertTrue(result.equals(new Complex(0,1)));
			
//            // Test whether a String is passed through
//            result = jep.evaluate(jep.parse("\"asdf\""));
//            Assert.assertTrue(result instanceof String);
//            Assert.assertEquals("asdf",result);
//    }
	
//    /**
//     * Test changing variables value after parsing
//     */
//    public void testChangeVariable()  throws Exception {
//        Object result;
//        Complex c;
//        // add the complex variable x = 0,0
//        jep.addVariable("x", new Complex(0, 0));
//        // parse a simple expression with the variable x
		
//        try {
//            jep.parse("x");
//        } catch (ParseException e) {
//            Assert.fail("Exception occured");
//        }
////		jep.getVarValue("x");
//        try {
//            result = jep.evaluate();
//        } catch (EvaluationException e) {
//            Assert.fail("Exception occured"+e.getMessage());
//            return;
//        }
//        Assert.assertTrue(result instanceof Complex);
//        c = (Complex)result;
//        Assert.assertTrue(c.re() == 0);
//        Assert.assertTrue(c.im() == 0);
//        // change the value of x
//        jep.addVariable("x", new Complex(1, 1));
////		jep.getVarValue("x");
//        try {
//            result = jep.evaluate();
//        } catch (EvaluationException e) {
//            Assert.fail("Exception occured");
//            return;
//        }
//        Assert.assertTrue(result instanceof Complex);
//        c = (Complex)result;
//        Assert.assertTrue(c.re() == 1);
//        Assert.assertTrue(c.im() == 1);
//    }
	

//    /**
//     * Tests whether allowUndeclared is working properly. 
//     *
//     */
//    public void testSetAllowUndeclared() throws Exception {
//        printTestHeader("Testing AllowedUndeclared options...");

//        // test whether setAllowUndeclared(true) works
//        jep.getVariableTable().clear();				// clear the Variable Table
//        jep.setAllowUndeclared(true);
//        try {
//            jep.parse("x");
//        } catch (ParseException e) {
//            Assert.fail("Exception occurred "+e.getMessage());
//        }
//        VariableTable st = jep.getVariableTable();
		
//        // should only contain a single variable x
//        Assert.assertTrue(st.size()==1);
//        Assert.assertTrue(st.getVariable("x") != null);
		
//        try {
//            jep.evaluate();
//            Assert.fail("Exception should have been thrown when trying to evaluate a declared but undefined variable");
//        } catch(EvaluationException e) {
//            // exception was thrown, so all is well
//        }
		
//        jep.setDefaultValue(Double.valueOf(0.0));
//        jep.parse("y");
//        Object val = jep.evaluate();
//        assertEquals("Value of y using default value",0.0,val);
		
//        // test whether setAllowUndeclared(false) works
//        jep.getVariableTable().clear();
//        jep.addVariable("x", new Double(1));
//        jep.setAllowUndeclared(false);
//        try {
//            jep.parse("p");
//            // since p is not declared, an error should occur
//            Assert.fail("A ParseException should have been thrown creating variable 'p'.");
//        } catch (ParseException e) {
//            // exception was thrown, so all is well
//        }
//    }


//    public void testNumbers() throws Exception
//    {
//        printTestHeader("Testing Numbers...");
	
//        valueTest("0",0.0);
//        valueTest("0.",0.0);
//        valueTest(".0",0.0);
//        valueTest("0.0",0.0);
//        valueTest("-0",-0.0);
//        valueTest("-0.0",-0.0);
//        valueTest("1",1.0);
//        valueTest("1.",1.0);
//        valueTest("1.0",1.0);
//        valueTest("-1",-1.0);
//        valueTest("-1.",-1.0);
//        valueTest("-1.0",-1.0);
//        valueTest("1.5",1.5);
//        valueTest(".5",0.5);
//        valueTest("0.5",0.5);
//        valueTest("-1.5",-1.5);
//        valueTest("-.5",-0.5);
//        valueTest("-0.5",-0.5);

//        valueTest("5e2",500.0);
//        valueTest("5E2",500.0);
//        valueTest("5e-2",0.05);
//        valueTest("5e-2",0.05);
//        valueTest("-5e2",-500.0);
//        valueTest("-5E2",-500.0);
//        valueTest("-5e-2",-0.05);
//        valueTest("-5e-2",-0.05);

//        valueTest(".5e3",500.0);
//        valueTest(".5E3",500.0);
//        valueTest(".5e-1",0.05);
//        valueTest(".5e-1",0.05);
//        valueTest("-.5e3",-500.0);
//        valueTest("-.5E3",-500.0);
//        valueTest("-.5e-1",-0.05);
//        valueTest("-.5e-1",-0.05);

//        valueTest("0.5e3",500.0);
//        valueTest("0.5E3",500.0);
//        valueTest("0.5e-1",0.05);
//        valueTest(".5e-1",0.05);
//        valueTest("-.5e3",-500.0);
//        valueTest("-.5E3",-500.0);
//        valueTest("-.5e-1",-0.05);
//        valueTest("-.5e-1",-0.05);

//        valueTest("1.5e3",1500.0);
//        valueTest("1.5E3",1500.0);
//        valueTest("1.5e-1",0.15);
//        valueTest("1.5e-1",0.15);
//        valueTest("-1.5e3",-1500.0);
//        valueTest("-1.5E3",-1500.0);
//        valueTest("-1.5e-1",-0.15);
//        valueTest("-1.5e-1",-0.15);

//    }
	
//    public void testStrings() throws Exception
//    {
//        printTestHeader("Testing strings...");

//        valueTest("\"\"","");
//        valueTest("\"a\"","a");
//        valueTest("\"abcdefghijklmnopqrstuvwxyz\"","abcdefghijklmnopqrstuvwxyz");
//        valueTest("sum(\"a\",\"b\")","ab");
//        valueTest("\"a\\\"b\"","a\"b");
//        valueTest("\"A\\bB\\fC\\nD\\rE\\tF\"","A\bB\fC\nD\rE\tF");
//        valueTest("A=\"1\"","1");
//        valueTest("B=\"2\"","2");
//        valueTest("(A==\"1\")&&(B==\"2\")",myTrue);
//    }

//    public void testLogical() throws Exception
//    {
//        printTestHeader("Testing logical operations");

//        valueTest("T=1",1.0);
//        valueTest("F=0",0.0);
//        valueTest("!T",myFalse);
//        valueTest("!F",myTrue);
//        valueTest("!5",myFalse);
//        valueTest("-0==0",myTrue);
//        valueTest("!-5",myFalse);
//        //valueTest("-!5==0",myTrue);
//        //valueTest("-!0",-1.0);
//        valueTest("T&&T",myTrue);
//        valueTest("T&&F",myFalse);
//        valueTest("F&&T",myFalse);
//        valueTest("F&&F",myFalse);
//        valueTest("T||T",myTrue);
//        valueTest("T||F",myTrue);
//        valueTest("F||T",myTrue);
//        valueTest("F||F",myFalse);
//        calcValue("a=F"); calcValue("b=F"); calcValue("c=F");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=F"); calcValue("b=F"); calcValue("c=T");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=F"); calcValue("b=T"); calcValue("c=F");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=F"); calcValue("b=T"); calcValue("c=T");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);

//        calcValue("a=T"); calcValue("b=F"); calcValue("c=F");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=T"); calcValue("b=F"); calcValue("c=T");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=T"); calcValue("b=T"); calcValue("c=F");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
//        calcValue("a=T"); calcValue("b=T"); calcValue("c=T");
//        valueTest("(a&&(b||c)) == ((a&&b)||(a&&c))",myTrue);
//        valueTest("(a||(b&&c)) == ((a||b)&&(a||c))",myTrue);
		
//        jep.getVariableTable().remove("true");
//        jep.getVariableTable().remove("false");
//        jep.addConstant("true",myTrue);
//        jep.addConstant("false",myFalse);
//        valueTest("true",myTrue);
//        valueTest("false",myFalse);
//        valueTest("!true",myFalse);
//        valueTest("!false",myTrue);
//        valueTest("true==true",myTrue);
//        valueTest("false==false",myTrue);
//        valueTest("true==false",myFalse);
//        valueTest("true==true&&false==false",myTrue);
//        valueTest("if(true==true&&false==false,6,7)",6.0);
//        valueTest("if(false&&true,6,7)",7.0);
//        valueTest("if(true&&false==false,6,7)",6.0);
//        valueTest("if((true&&true)==true,6,7)",6.0);
//        valueTest("if((!false)==true,6,7)",6.0);
//    }

//    protected Object myTrue = Boolean.TRUE; 
//    protected Object myFalse = Boolean.FALSE; 
//    public void testNaN() throws Exception
//    {
//        printTestHeader("Testing for NaN");
//        double x=1.0,y=0.0;
//        double z=x/y;
//        if(z<2.0)
//            System.out.println("1/0<2 is true");
//        else
//            System.out.println("1/0<2 is false");
		
//        jep.addVariable("x",new Double(Double.NaN));
//        System.out.println("x=NaN");
//        valueTestNaN("ln(x)");
//        valueTestNaN("log(x)");
//        valueTestNaN("sin(x)");
//        valueTestNaN("x+x");
//        valueTest("x!=x",myTrue);
//        valueTest("x==x",myFalse);

//        jep.addVariable("y",new Double(Double.NaN));
//        Node n = jep.parse("x+5");
//        System.out.println(calcValue(n));
//        Node n2 = jep.parse("y");
//        System.out.println(calcValue(n2));
//        valueTest("x == x+5",myFalse);
//        valueTest("x == 0/0",myFalse);
//        valueTest("x == x",myFalse);
//        valueTest("x == 0 * x",myFalse);
//        valueTest("x == 5",myFalse);
//        valueTest("x == y",myFalse);
//        valueTest("y == y",myFalse);
//        System.out.println("Set x to Double(5)");
//        jep.addVariable("x",new Double(5));
//        valueTest("x == x+5",myFalse);
//        valueTest("x == x",myTrue);
//    }

//    public void testComplex() throws Exception
//    {
//        printTestHeader("Testing complex values");
//        double tol = 0.00000001;

//        complexValueTest("z=complex(3,2)",new Complex(3,2),tol);
//        complexValueTest("z*z-z",new Complex(2,10),tol);
//        complexValueTest("z^3",new Complex(-9,46),tol);
//        complexValueTest("(z*z-z)/z",new Complex(2,2),tol);
//        complexValueTest("w=polar(2,pi/2)",new Complex(0,2),tol);
		
//        complexValueTest("ln(-1)",new Complex(0,Math.PI),tol);
//        complexValueTest("sqrt(-1)",new Complex(0,1),tol);
//        complexValueTest("pow(-1,0.5)",new Complex(0,1),tol);
//        valueTest("arg(w)",Math.PI/2);
//        valueTest("cmod(w)",2.0);
//        valueTest("re(z)",3.0);
//        valueTest("im(z)",2.0);
//        complexValueTest("conj(z)",new Complex(3,-2),tol);
//        complexValueTest("exp(pi*i/2)",new Complex(0,1),tol);
//        //complexValueTest("cos(z)",new Complex(3,-2),tol);
//    }

//    public void testFunction() throws Exception
//    {
//        printTestHeader("Testing real functions");
//        valueTest("abs(2.5)",2.5);
//        valueTest("abs(-2.5)",2.5);
//        valueTest("acos(1/sqrt(2))",Math.PI/4,0.00000001);
//        valueTest("cos(pi/3)",0.5,0.00000001);
//        valueTest("atan2(3,4)",Math.atan2(3.0,4.0),0.00000001);
//        valueTest("2^4",16.0);
//        valueTest("2^10",1024.0);
//        valueTest("2^0.5",Math.sqrt(2),0.00000001);
//        valueTest("2^(-0.5)",1.0/Math.sqrt(2),0.00000001);
//        valueTest("2^(-4)",1.0/16,0.00000001);
//        valueTest("2^(-10)",1.0/1024,0.00000001);
//        calcValue("rand()");
//        valueTest("4 atan2(1,1)",Math.PI,0.00000001);
//    }

//    public void testIf()  throws Exception
//    {
//        printTestHeader("Testing if statement");
//        valueTest("if(1,2,3)",2.0);		
//        valueTest("if(-1,2,3)",3.0);		
//        valueTest("if(0,2,3)",3.0);		
//        valueTest("if(1,2,3,4)",2.0);		
//        valueTest("if(-1,2,3,4)",3.0);		
//        valueTest("if(0,2,3,4)",4.0);		
//        valueTest("if(0>=0,2,3,4)",2.0);		
//        valueTest("x=3",3.0);		
//        valueTest("if(x==3,1,-1)",1.0);		
//        valueTest("if(x!=3,1,-1)",-1.0);		
//        valueTest("if(x>=3,1,-1)",1.0);		
//        valueTest("if(x>3,1,-1)",-1.0);		
//        valueTest("if(x<=3,1,-1)",1.0);		
//        valueTest("if(x<3,1,-1)",-1.0);		
//    }

//    public void testPlusPlus() throws Exception
//    {
//        printTestHeader("++ notation");
//        valueTest("3++2",5.0);
//        valueTest("3+-2",1.0);
//        valueTest("3-+2",1.0);
//        valueTest("3--2",5.0);
//        valueTest("3+++2",5.0);
//        valueTest("3++-2",1.0);
//        valueTest("3+-+2",1.0);
//        valueTest("3+--2",5.0);
//        valueTest("3-++2",1.0);
//        valueTest("3-+-2",5.0);
//        valueTest("3--+2",5.0);
//        valueTest("3---2",1.0);

//    }
	
//    public void testImplicitMul() throws Exception
//    {
//        printTestHeader("Implicit Multiplication");
//        valueTest("x=5",5.0);
//        valueTest("y=7",7.0);
//        valueTest("x2=37",37.0);
//        valueTest("2 x",10.0);
//        valueTest("2x",10.0);
//        valueTest("x 2",10.0);
//        valueTest("x2",37.0);
//        valueTest("-2 x",-10.0);
//        valueTest("x y",35.0);
//        valueTest("-x y",-35.0);
//        valueTest("x x x x x x x x",Math.pow(5, 8));
//        valueTest("3 (2+4)",18.0);
//        valueTest("3(2+4)",18.0);
//        valueTest("2(x-1)",8.0);
//        valueTest("(x+1)(x-1)",24.0);
//        valueTest("x(x-1)",20.0);
//        valueTest("(x+1)x",30.0);
//        valueTest("1+2 3",7.0);
//        valueTest("2 3+4",10.0);
//        valueTest("1-2 3",-5.0);
//        valueTest("2 3 4",24.0);
//        valueTest("-(4-2)",-2.0);
//    }
	
//    public void testUminusPower() throws Exception
//    {
//        printTestHeader("Testing precedence of unitary minus and power...");
//        valueTest("-1^2", -1.0);
//        valueTest("2^-3", 0.125);
//    }
	
//    public void testNumParam() throws Exception
//    {
//        printTestHeader("Number of parameters");
//        jep.parse("if(3,1,2)");
//        jep.parse("if(4,1,2,3)");
//        try
//        {
//            jep.parse("if(5,1,2,3,4)");
//            fail("Did not trap illegal number of arguments");
//        }
//        catch(ParseException e) {/* ignore */ }
//        jep.parse("a1=1234");
//        jep.parse("a2=5678");
//        jep.parse("ApportionmentAmt=4321");
//        jep.parse("a4 = 2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + + 1200");
//        jep.parse("a3 = if(a1 > 0 && ApportionmentAmt < 1000, if(a2 < 2000, if(a2 < 1000, 200, 0), if(a1 > 1000, if((2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + 1200 + ApportionmentAmt / 2000 + ApportionmentAmt * ApportionmentAmt + 2000) > 0, 100, 200),200)), if(a1/a2 < 1000, a1/a2, 1, a1 * a2 + a1))");
//        try
//        {
//         jep.parse("a3 = if(a1 > 0 && ApportionmentAmt < 1000, if(a2 < 2000, if(a2 < 1000, 200, 0), if(a1 > 1000, if((2000 + (3000 /2000) + (3.45787 * 33544 - (212.223 /2000)) + 1200 + ApportionmentAmt / 2000 + ApportionmentAmt * ApportionmentAmt + 2000) > 0, 100, 200)),200), if(a1/a2 < 1000, a1/a2, 1, a1 * a2 + a1))");
//            fail("Did not trap illegal number of arguments");
//        }
//        catch(ParseException e) { /* ignore */}
//    }

//    public void testBinom() throws ParseException,Exception
//    {
//        printTestHeader("Testing binomial coeffs");
//        valueTest("binom(0,0)",1);
//        valueTest("binom(1,0)",1);
//        valueTest("binom(1,1)",1);
//        valueTest("binom(2,0)",1);
//        valueTest("binom(2,1)",2);
//        valueTest("binom(2,2)",1);
//        valueTest("binom(3,0)",1);
//        valueTest("binom(3,1)",3);
//        valueTest("binom(3,2)",3);
//        valueTest("binom(3,3)",1);
//        valueTest("binom(4,0)",1);
//        valueTest("binom(4,1)",4);
//        valueTest("binom(4,2)",6);
//        valueTest("binom(4,3)",4);
//        valueTest("binom(4,4)",1);
//        valueTest("binom(5,0)",1);
//        valueTest("binom(5,1)",5);
//        valueTest("binom(5,2)",10);
//        valueTest("binom(5,3)",10);
//        valueTest("binom(5,4)",5);
//        valueTest("binom(5,5)",1);

//        valueTest("binom(6,0)",1);
//        valueTest("binom(6,1)",6);
//        valueTest("binom(6,2)",15);
//        valueTest("binom(6,3)",20);
//        valueTest("binom(6,4)",15);
//        valueTest("binom(6,5)",6);
//        valueTest("binom(6,6)",1);
		
//        valueTest("binom(10,1)",10);
//        valueTest("binom(10,5)",252);
//    }

//    public void testFormat() throws Exception
//    {
//        printTestHeader("Format");
//        NumberFormat format = NumberFormat.getInstance();
//        jep.getPrintVisitor().setNumberFormat(format);
//        format.setMaximumFractionDigits(3);
//        format.setMinimumFractionDigits(0);
		
//        String s1 = "[10,0,0.1,0.11,0.111,0.1111]";
//        String r1 = jep.toString(jep.parse(s1));
//        this.myAssertEquals(s1,"[10,0,0.1,0.11,0.111,0.111]",r1);
//        String s2 = "[0.9,0.99,0.999,0.9999]";
//        String r2 = jep.toString(jep.parse(s2));
//        this.myAssertEquals(s2,"[0.9,0.99,0.999,1]",r2);
		
//        String s3 = "8 - 7.9"; 
//        jep.parse(s3);
//        double res = jep.evaluateD();
//        String s = String.format("%.3f",res);
//        this.myAssertEquals(s3,"0.100",s);
		
//        String s4 = "round(8-7.9,3)";
//        jep.parse(s4);
//        res = jep.evaluateD();
//        this.myAssertEquals(s4,"0.1",String.valueOf(res));
//    }

//    public void testAssign()  throws Exception
//    {
//        printTestHeader("Assignment of variables");
//        valueTest("x=3",3.0);
//        valueTest("y=3+4",7.0);
//        valueTest("z=x+y",10.0);
//        valueTest("a=b=c=z",10.0);
//        valueTest("b",10.0);
//        valueTest("d=f=a-b",0.0);
//        valueTest("x=2",2.0);
//        valueTest("(x*x)*x*(x*x)",32.0); // Works fine with Multiply
////		new org.lsmp.djep.vectorJep.VectorJep();
//        valueTest("(x*x)*x*(x*x)",32.0);
//        // this created an error in 2.3.0b
//        // as creating a VectorJep changed the operator set
//        // and hence the broken MMultiply was used.								
//    }

//    public void testNoAssign()  throws Exception
//    {
//        printTestHeader("Assignment of variables");
//            jep.setAllowAssignment(false);
//            try
//            {
//                jep.parse("x=3");
//                fail("Assignment should not have been possible");
//            }
//            catch(ParseException e) {
//                System.out.println("Attept at assignment sucessfully caught");
//            }
//            jep.setAllowAssignment(true);
//            jep.parse("x=3");
//    }

//    public void testListAccess() throws Exception
//    {
//        printTestHeader("List operations");
//        valueTestString("x=[4,3,2,1]","[4.0, 3.0, 2.0, 1.0]");
//        valueTest("x[2]",3.0);
//        valueTest("x[4]=5",5.0);
//        valueTestString("x","[4.0, 3.0, 2.0, 5.0]");
//        valueTestString("[]","[]");
//        valueTestString("x+x","[8.0, 6.0, 4.0, 10.0]");
//        valueTestString("x-x","[0.0, 0.0, 0.0, 0.0]");
//        valueTestString("x*2","[8.0, 6.0, 4.0, 10.0]");
//        valueTestString("2*x","[8.0, 6.0, 4.0, 10.0]");
//        valueTestString("x/2","[2.0, 1.5, 1.0, 2.5]");
//        valueTestString("w=[1+i,1-i]","[(1.0, 1.0), (1.0, -1.0)]");
//        valueTestString("2*w","[(2.0, 2.0), (2.0, -2.0)]");
//    }

//    public void testMultiDimArray() throws Exception
//    {
//        printTestHeader("List operations");
//        valueTestString("x=[[1,2],[3,4]]","[[1.0, 2.0], [3.0, 4.0]]");
//        valueTestString("x[2]","[3.0, 4.0]");
//        valueTestString("x+x","[[2.0, 4.0], [6.0, 8.0]]");
//        valueTestString("x-x","[[0.0, 0.0], [0.0, 0.0]]");
//        valueTestString("x*2","[[2.0, 4.0], [6.0, 8.0]]");
//        valueTestString("2*x","[[2.0, 4.0], [6.0, 8.0]]");
//        valueTestString("x/2","[[0.5, 1.0], [1.5, 2.0]]");
//    }

//    public void testLazyLogical() throws Exception
//    {
//        printTestHeader("Lazy Logical");
//        jep.getOperatorTable().getAnd().setPFMC(new LazyLogical(LazyLogical.AND));
//        jep.getOperatorTable().getOr().setPFMC(new LazyLogical(LazyLogical.OR));
//        class SideEffect extends PostfixMathCommand {
//        private static final long serialVersionUID = 1L;
//        public SideEffect() { this.numberOfParameters = 1; }
//        public boolean called = false;
//        public void run(Stack<Object> aStack) throws EvaluationException {
//            called = true;
//        }
//        }
//        SideEffect se = new SideEffect();
//        jep.addFunction("sideEffect",se);
//        Node n1 = jep.parse("0 || sideEffect(1==2)");
//        Object v1 = jep.evaluate(n1);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(myFalse,v1);
//        se.called = false;

//        Node n2= jep.parse("1 || sideEffect(1==2)");
//        Object v2 = jep.evaluate(n2);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(myTrue,v2);
//        se.called = false;

//        Node n3 = jep.parse("0 && sideEffect(1==2)");
//        Object v3 = jep.evaluate(n3);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(myFalse,v3);
//        se.called = false;

//        Node n4= jep.parse("1 && sideEffect(1==2)");
//        Object v4 = jep.evaluate(n4);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(myFalse,v4);
//        se.called = false;

//        Node n5 = jep.parse("0 || sideEffect(2==2)");
//        Object v5 = jep.evaluate(n5);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(myTrue,v5);
//        se.called = false;

//        Node n6= jep.parse("1 || sideEffect(2==2)");
//        Object v6 = jep.evaluate(n6);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(myTrue,v6);
//        se.called = false;

//        Node n7 = jep.parse("0 && sideEffect(2==2)");
//        Object v7 = jep.evaluate(n7);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(myFalse,v7);
//        se.called = false;

//        Node n8= jep.parse("1 && sideEffect(2==2)");
//        Object v8 = jep.evaluate(n8);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(myTrue,v8);
//        se.called = false;

//        Node n9= jep.parse("if(1,sideEffect(2),3)");
//        Object v9 = jep.evaluate(n9);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(new Double(2),v9);
//        se.called = false;

//        Node n10= jep.parse("if(1,2,sideEffect(3))");
//        Object v10 = jep.evaluate(n10);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(new Double(2),v10);
//        se.called = false;

//        Node n11= jep.parse("if(0,sideEffect(2),3)");
//        Object v11 = jep.evaluate(n11);
//        Assert.assertFalse(se.called);
//        Assert.assertEquals(new Double(3),v11);
//        se.called = false;

//        Node n12= jep.parse("if(0,2,sideEffect(3))");
//        Object v12 = jep.evaluate(n12);
//        Assert.assertTrue(se.called);
//        Assert.assertEquals(new Double(3),v12);
//        se.called = false;
//    }
	
//    public void testLogarithm() throws Exception
//    {
//        printTestHeader("Logarithm");
//        FunctionTable ft = jep.getFunctionTable();
//        ft.addFunction("slog",new StrictNaturalLogarithm());
//        ft.addFunction("slog10",new StrictNaturalLogarithm(10));
//        ft.addFunction("slog2",new StrictNaturalLogarithm(2));
//        jep.addVariable("NaN",Double.NaN);
//        valueTest("ln(1)",0.0,1E-7);
//        valueTest("slog(1)",0.0,1E-7);
//        valueTest("slog10(1)",0.0,1E-7);
//        valueTest("ln(e)",1.0,1E-7);
//        valueTest("slog(e)",1.0,1E-7);
//        valueTest("slog10(10)",1.0,1E-7);
//        valueTestNaN("slog(-1)");
//        valueTestNaN("slog10(-1)");
//        valueTestNaN("slog2(-1)");
//    }
	
//    /**
//     * Test creating a Jep instance with no functions and no variables.
//     * @throws Exception
//     */
//    public void testBlankParser() throws Exception
//    {
//        printTestHeader("Testing blank parser");
//        Jep j = new Jep();
//        // override standard variable table with blank one
//        j.setComponent(new VariableTable());
//        // override standard function table with blank one
//        j.setComponent(new FunctionTable());
		
//        // ensure function and variable tables are blank
//        Assert.assertTrue(j.getFunctionTable().size() == 0);
//        Assert.assertTrue(j.getVariableTable().size() == 0);
//        // ensure operators still work
//        j.parse("1+1");
//        Object res = j.evaluate();
//        Assert.assertTrue(res instanceof Double);
//        Assert.assertTrue(((Double)res).doubleValue()==2.0);
//    }
	
//    public void testMacroFunction() throws Exception
//    {
//        printTestHeader("Testing MacroFunctions...");
//        MacroFunction fact = new MacroFunction(
//            "fact",new String[]{"x"}, 
//            "if(x>1,x*fact(x-1),1)");
//        jep.addFunction("fact",fact);
//        fact.init(jep);
//        valueTest("fact(5)",120.0);
//        jep.addVariable("x",4);
//        valueTest("fact(5)",120.0);
//        Assert.assertEquals(4.0,jep.getVariableValue("x"));
//        valueTest("x+fact(x+1)",124.0);
	    
//    }
	
//    /**
//     * Tests whether a parse exceptions are thrown when they should be. 
//     * See ticket #50
//     * Tests
//     * on both the standard parser and the configurable parser, is carried out
//     * by CPTest which extends this.
//     */
//    public void testParseException() {
//        printTestHeader("Testing ParseException...");

//        // all these expressions should cause ParseException
//        String expr[] = {"1+", "(", ")", "sin(", "(1+)", "*1", "1*", "-","avg()","min()","max()"};
		
//        // for each expression
//        for (int i=0;i<expr.length;i++) {
//            try {
//                System.out.println("Attempting parse of \"" + expr[i]+ "\"");
//                jep.parse(expr[i]);
//                // expression should have failed to parse
//                Assert.fail("Attempting parse of \"" + expr[i]+ "\"");
//            } catch (ParseException e) {
//                // expected a parse exception has been thrown
//                System.out.println("Caught expected exception \"" + e.getMessage()+"\"");
//            } catch (Exception e) {
//                // some other exception was thrown
//                Assert.fail("Attempting parse of \"" + expr[i]+ "\"" + e.getMessage());
//            }
//        }
//    }

//    /**
//     * Test whether evaluation exceptions are thrown when they should be.
//     */
//    public void testEvalExceptions() {
//        printTestHeader("Ensuring evaluation exception are thrown when they should be...");
//        ArrayList<String> exprs = new ArrayList<String>();

//        // random expressions that will fail during evaluation
//        //exprs.add("avg()");
//        //exprs.add("min()");
//        //exprs.add("max()");
//        exprs.add("avg([])");
//        exprs.add("min([])");
//        exprs.add("max([])");

//        // add expressions like sin([1, 1]) that can not handle arrays
//        String noArrayFuncs[] = { // list of functions that can not take arrays as arguments
//                "sin", "cos", "tan", "asin", "acos", "atan", 
//                "sinh", "cosh", "tanh", "asinh", "acosh", "atanh",
//                "log", "ln", "exp"
//        };
//        for (String funcname:noArrayFuncs) {
//            exprs.add(funcname+"([])");
//            exprs.add(funcname+"([1,1])");			
//        }
		
//        // loop through each exception
//        for (String str:exprs) {
//            // parsing should work fine
//            try {
//                jep.parse(str);
//            } catch (ParseException e) {
//                Assert.fail("Could not parse expression: \"" + str + "\", "+e.getMessage());
//            }
//            // but evaluation should fail
//            try {
//                jep.evaluate();
//                // we shouldn't get here because these expressions should
//                // end up throwing eval expections
//                Assert.fail("No evaluation exception was thrown as expected.");
//            } catch (EvaluationException e) {
//                // we should end up here
//            }
//        }
//    }
	
//    public void testSemiColon() throws Exception
//    {
//     String s = "x=5; y=6; x+y";
//    printTestHeader("Testing reading multiple equations: "+s);
//     jep.initMultiParse(s);
//     Node n1 = jep.continueParsing();
//     Node n2 = jep.continueParsing();
//     Node n3 = jep.continueParsing();
//     Object o1 = jep.evaluate(n1);
//     this.myAssertEquals("x=5", 5.0,o1);
//     Object o2 = jep.evaluate(n2);
//     this.myAssertEquals("y=6", 6.0,o2);
//     Object o3 = jep.evaluate(n3);
//     this.myAssertEquals("x+y", 11.0,o3);
//     Node n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);
//    }

//    public void testMultiLine() throws Exception
//    {
//     String s = "\nx=\n5;y=6;\nx+y";
//    printTestHeader("Testing reading multiple equations: "+s);
//     jep.initMultiParse(s);
	 
//     Node n1 = jep.continueParsing();
//     Object o1 = jep.evaluate(n1);
//     this.myAssertEquals("x=5", 5.0,o1);

//     Node n2 = jep.continueParsing();
//     Object o2 = jep.evaluate(n2);
//     this.myAssertEquals("y=6", 6.0,o2);

//     Node n3 = jep.continueParsing();
//     Object o3 = jep.evaluate(n3);
//     this.myAssertEquals("x+y", 11.0,o3);

//     Node n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);
	 
//     /******/
	 
//     s = "x*y";
//     jep.initMultiParse(s);
	 
//     n1 = jep.continueParsing();
//     o1 = jep.evaluate(n1);
//     myAssertEquals("x*y", 30.0,o1);
	 
//     n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);

//     /******/

//     s = "\n";
//     jep.initMultiParse(s);
//     n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);

//     /******/

//     s = "";
//     jep.initMultiParse(s);
//     n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);

//     /******/

//     s = "\n\n\n";
//     jep.initMultiParse(s);
//     n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);

//     /******/

//     s = "\n\n\n1\n\n\n";
//     jep.initMultiParse(s);
//     n1 = jep.continueParsing();
//     o1 = jep.evaluate(n1);
//     myAssertEquals("1", 1.0,o1);
//     n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);

//    }


//    public void testEmptyEqn() throws Exception
//    {
//     String s = "x=5;;y=6;x+y";
//     printTestHeader("Testing reading multiple equations: "+s);
//     jep.initMultiParse(s);

//     Node n1 = jep.continueParsing();
//     Object o1 = jep.evaluate(n1);
//     this.myAssertEquals("x=5", 5.0,o1);

//     Node n2 = jep.continueParsing();
//     Object o2 = jep.evaluate(n2);
//     this.myAssertEquals("y=6", 6.0,o2);

//     Node n3 = jep.continueParsing();
//     Object o3 = jep.evaluate(n3);
//     this.myAssertEquals("x+y", 11.0,o3);

//     Node n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);
//    }

//    public void testComments() throws Exception
//    {
//        printTestHeader("Testing comments: ");
//        NodeFactory nf = jep.getNodeFactory();
//        OperatorTable ot = jep.getOperatorTable();

//        Node expected = 
//        nf.buildOperatorNode(
//            ot.getDivide(), 
//            nf.buildOperatorNode(
//                ot.getMultiply(),
//                nf.buildVariableNode("x"),
//                nf.buildVariableNode("y")),
//            nf.buildVariableNode("z"));

//        String s = "x*y/z //foo\n";
//        Node node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "//foo\nx*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x//foo\n*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* foo */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* foo */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/ /* foo */z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/* foo *//z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*/* foo */y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x/* foo */*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "/* foo */x*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);


	    
//        s = "x*y/z /* foo/bar */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z/* foo*bar */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

	    
//        s = "x*y/z/* foo\n */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/ /* foo\n */z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/* foo\n *//z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*/* foo\n */y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x/* foo\n */*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "/* foo\n */x*y/z";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* foo/bar\n */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* foo*bar\n */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);


//        s = "x*y/z /* \nfoo/bar */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* \nfoo*bar */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* \nfoo/bar\n */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//        s = "x*y/z /* \nfoo*bar\n */";
//        node = jep.parse(s);
//        this.assertEqNodes(s, expected,node);

//    }

//    public void testEmptyEqn2() throws Exception
//    {
//     String s = "x=5;/*foo*/;y=6;x+y";
//     printTestHeader("Testing reading multiple equations: "+s);
//     jep.initMultiParse(s);

//     Node n1 = jep.continueParsing();
//     Object o1 = jep.evaluate(n1);
//     this.myAssertEquals("x=5", 5.0,o1);

//     Node n2 = jep.continueParsing();
//     Object o2 = jep.evaluate(n2);
//     this.myAssertEquals("y=6", 6.0,o2);

//     Node n3 = jep.continueParsing();
//     Object o3 = jep.evaluate(n3);
//     this.myAssertEquals("x+y", 11.0,o3);

//     Node n4 = jep.continueParsing();
//     assertNull("Next parse should give a null result",n4);
//    }

//    public void testX2Y() throws Exception
//    {
//        String s= "x^2y";
//        Node node = jep.parse(s);
//        NodeFactory nf = jep.getNodeFactory();
//        OperatorTable ot = jep.getOperatorTable();
//        Node expected = 
//        nf.buildOperatorNode(
//            ot.getMultiply(), 
//            nf.buildOperatorNode(
//                ot.getPower(),
//                nf.buildVariableNode("x"),
//                nf.buildConstantNode(2.0)),
//            nf.buildVariableNode("y"));
//        this.assertEqNodes(s, expected,node);
	    
//        s= "!x y";
//        node = jep.parse(s);
//        expected = 
//        nf.buildOperatorNode(
//            ot.getMultiply(), 
//            nf.buildOperatorNode(
//                ot.getNot(),
//                nf.buildVariableNode("x")),
//            nf.buildVariableNode("y"));
//        this.assertEqNodes(s, expected,node);

//        s= "x*2y";
//        node = jep.parse(s);
//        expected = 
//        nf.buildOperatorNode(
//            ot.getMultiply(), 
//            nf.buildOperatorNode(
//                ot.getMultiply(),
//                nf.buildVariableNode("x"),
//                nf.buildConstantNode(2.0)),
//            nf.buildVariableNode("y"));
//        this.assertEqNodes(s, expected,node);

//        s= "x/2y";
//        node = jep.parse(s);
//        expected = 
//        nf.buildOperatorNode(
//            ot.getMultiply(), 
//            nf.buildOperatorNode(
//                ot.getDivide(),
//                nf.buildVariableNode("x"),
//                nf.buildConstantNode(2.0)),
//            nf.buildVariableNode("y"));
//        this.assertEqNodes(s, expected,node);

//        s= "x+2y";
//        node = jep.parse(s);
//        expected = 
//        nf.buildOperatorNode(
//            ot.getAdd(),
//            nf.buildVariableNode("x"),
//            nf.buildOperatorNode(
//                ot.getMultiply(),
//                nf.buildConstantNode(2.0),
//                nf.buildVariableNode("y")));
//        this.assertEqNodes(s, expected,node);

//    }
	
//    /**
//     * Tests for bug #52
//     */
//    public void testExpectedTokenSequence() {
//        printTestHeader("Checking whether expected token sequence is set");
//        Jep j = new Jep();
//        try {
//            j.parse("1+");
//        } catch (com.singularsys.jep.ParseException e) {
//            Throwable t = e.getCause();
//            // ensure the cause is set
//            assertNotNull(t);
//            if (t instanceof com.singularsys.jep.parser.ParseException) {
//                com.singularsys.jep.parser.ParseException specialE = 
//                    (com.singularsys.jep.parser.ParseException)t;
//                System.out.println("Contains parser.ParseException :)");
//                int expectedTokens[][] = specialE.expectedTokenSequences;
//                Assert.assertTrue(expectedTokens.length > 0);
//            } else {
//                Assert.fail();
//            }
//        }	
//    }
	
//    /**
//     * Test for bug with 1&&1&&1
//     * 
//     */
	
//    public void testLazyLogicalBug() throws Exception {
//        valueTest("true&&1",myTrue);
//    }
//}
#endregion